
#
# LM-MA-ES
# Copyright (c) 2017 Tobias Glasmachers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import numpy as np
import math
import sys
import time


def sphere(x):
	return np.dot(x, x)

elli_factors = np.zeros(0)
def elli(x):
	global elli_factors
	if elli_factors.shape != x.shape:
		dim = x.shape[0]
		elli_factors = np.zeros(dim)
		for i in range(dim):
			elli_factors[i] = math.pow(1e-6, i / (dim - 1))
	return np.dot(x, elli_factors * x)

def rosen(x):
	n = len(x)
	return np.sum(100.0 * np.square(x[1:n] - np.square(x[0:n-1])) + np.square(x[0:n-1] - np.ones(n-1)))




def report(fe, fitness, sigma):
	sys.stdout.write('\b' * 100);
	sys.stdout.write("            #FE: " + str(fe) + "  fitness: " + str(fitness) + "  step size: " + str(sigma) + "     \b\b\b\b\b")
	sys.stdout.flush()

def lm_ma_es(x, sigma, f, f_target, budget = 1000000, verbosity = 1):
	# preparation
	start = time.time()
	n = len(x)
	m = int(4 + math.floor(3 * math.log(n)))
	lbd = m
	mu = lbd // 2
	assert lbd < n     # otherwise tuning constants break - use standard CMA-ES instead :)
	w = [math.log(lbd/2 + 0.5) - math.log(i + 1) for i in range(mu)]
	w /= np.sum(w)
	mu_w = 1 / np.sum(np.square(w))
	c_sigma = 2 * lbd / n
	c_sigma_update = math.sqrt(mu_w * c_sigma * (2 - c_sigma))
	c_d = np.zeros((m,))
	c_c = np.zeros((m,))
	c_c_update = np.zeros((m,))
	for i in range(m):
		c_d[i] = 1 / (math.pow(1.5, i) * n)
		c_c[i] = lbd / (math.pow(4.0, i) * n)
		c_c_update[i] = math.sqrt(mu_w * c_c[i] * (2 - c_c[i]))
	d_sigma = 1.0
	p_sigma = np.zeros((n,))
	M = np.zeros((m, n))
	fd = np.zeros((lbd,))
	chi = math.sqrt(n) * (1 - 1/(4*n) - 1/(21*n*n));

	if verbosity >= 1:
		print("[LM-MA-ES]  n=" + str(n) + "  m=" + str(m) + "  mu=" + str(mu) + "  lambda=" + str(lbd))

	# generation loop
	t = 0
	while True:
		# sample offspring, vectorized version
		z = np.random.randn(lbd, n)
		d = np.copy(z)
		for j in range(min(t, m)):
			d = (1 - c_d[j]) * d + c_d[j] * np.outer(np.dot(d, M[j,:]), M[j,:])

		# evaluate offspring and check stopping criterion
		for i in range(lbd):
			if t * lbd + i >= budget:
				report(budget, fd[order[0]], sigma)
				print("  budget exceeded; (" + str(time.time() - start) + " seconds)")
				return x, budget
			fd[i] = f(x + sigma * d[i,:])
			if fd[i] <= f_target:
				report(t * lbd, fd[i], sigma)
				print("  target fitness reached (" + str(time.time() - start) + " seconds)")
				return x + sigma * d[i,:], t * lbd + i + 1

		# sort by fitness
		order = np.argsort(fd)

		# update mean
		for i in range(mu):
			x += sigma * w[i] * d[order[i],:]

		# compute weighted mean
		wz = np.zeros((n,))
		for i in range(mu):
			wz += w[i] * z[order[i],:]

		# update evolution path
		p_sigma *= 1 - c_sigma
		p_sigma += c_sigma_update * wz

		# update direction vectors
		for i in range(m):
			M[i,:] *= 1 - c_c[i]
			M[i,:] += c_c_update[i] * wz

		# update step size
		sigma *= math.exp(c_sigma / d_sigma * (np.linalg.norm(p_sigma) / chi - 1))

		# status report
#		if verbosity >= 1 and t % 1000 == 0:
#			report(t * lbd, fd[order[0]], sigma)
		report(t * lbd, fd[order[0]], sigma)

		# generation counter
		t += 1


# configuration
f = rosen
dim = 1000

# experiment
np.random.seed(1001)
x, t = lm_ma_es(np.random.randn(dim), 1.0, f, 1e-10, 20 * dim)
